package audio;

import java.util.*;
import java.io.*;

/* the audio package is an interface to audio cards on the machine.
 * it provides sound input and output capabilities, and can manage
 * multiple sound cards simultaneously
 *  use of this system is limited in many ways, and is as follows:
 *  dev = new Audio(devID);
 *  dev.in.playBytes(...)
 *  dev.out.recBytes(...)
 *  dev.in.stop(..)
 * etc.
 *
 * Ok, there is a lot of functionality that could be made, and only some
 * that I really want.
 * So, I am going to redesign this, hopefully more simply and easier.
 * We need:
 *   incomming sample events (fine grained)
 *   outgoing sample finished events (coarse grained)
 *     - these can be just vectors of fine grained events, together with a mask field
 *   you would like to be able to output the same block many times, possibly simultaneously
 *   incomming blocks obviously can not be output, nor input, other than the original action
 *
 * So, what we do is the following:
 * when a block is sent for input, it is marked as 'used for input' and put on sent-queue
 * when a block is recvd from input, it is marked as 'not in use', taken from sent-queue
 * when a block is sent to output, a counter is incremented 'in use for output' and put on sent-queue
 * when a block is recvd from output, the counter is decremented, taken from sent-queue
 * this is basically the reader-writer problem, now that I step back.
 *
 * when a block has no more users, the internal buffers can be freed
 * also, the block can't be modified (data, len, notification attribute, etc) while it is in-use
 *
 * there are no 'done queues' anymore. the app can do that themselves. We notify the block when it finishes.
 * block.doneInput();
 * but, we want to have more than one thread sending the same block, but getting back the correct event when
 * the correct instance finishes. We maybe can do this with some kind of cookies?
 * send a block for input, then wait on the block to be finished
 * send a block for output, get a cookie back (what number in line are you), and wait on block with the specified cookie
 * no, this is stupid, you should really just make a new audio block with the same data.
 * send a block for output, then wait on it to be finished
 * we still need the sent-queues, though, so that the pointers dont dissapear....

  *
  *
  * So, we now have that blocks are not modified in a bad way.
 * now, for the notification...
 * in order to get an incomming sample event, we can wait on the input-lock
 *  when a block is recvd, it goes through
	  synch(input-lock) { sent-queue.get(); rcvd-queue.put(..), input-lock.notify() } // only notify ONE thread....
 *  and so we can wait by using
 *    synch(input-lock) { if (empty) input-lock.wait(); block = rcvd-queue.get(); } 
 * in order to get an outgoing-finished event, we can wait on the output-lock
 * same dealm except that output blocks are not kept around. So, we really want to do a block.waitTilFinished() call
 * this is easy...
  
	*   a Sample is a collection of AudioBlocks
 *   In order to input, you have to feed audio blocks to the device, 
 *     and you get back 
 */

public class Audio {

	public static boolean debug = true;
	public static void println(String s) { if (debug) System.out.println("Audio: " + s); }

	// global device info
	static private int numDev;
	static protected String deviceName[];
	static protected Audio devices[];
	static protected Hashtable hwoDevices = new Hashtable();
	static protected Hashtable hwiDevices = new Hashtable();
	// global message thread handle
	static private AudioMessageThread messageThread;

	// static initializer for devices and message thread
	static {
		try {
			System.loadLibrary("jkaudio");
			numDev = getNumDevicesN();
			deviceName = new String[numDev];
			devices = new Audio[numDev];
			// how to assign names to the devices?
			for(int i = 0; i < numDev; i++) {
				deviceName[i] = "Sound Card "+i;
				devices[i] = null;
			}
		} catch (Error e) {
			e.printStackTrace();
			throw e;
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	// global information accessors
	private static native int getNumDevicesN();

	public static int getNumDevices() {return numDev;}

	public boolean inUse(int devID) {
		return devices[devID] != null;
	}
	
	public static int inUse() {
		int c = 0;
		for(int i = 0; i < numDev; i++) {
			if (devices[i] != null) c++;
		}
		return c;
	}

	public static Audio getDeviceByHwo(int hwo) {
		return (Audio)hwoDevices.get(new Integer(hwo));
	}
	public static Audio getDeviceByHwi(int hwi) {
		return (Audio)hwiDevices.get(new Integer(hwi));
	}

	// Per-Device information
	public int devID;
	public String devName; 
	public  AudioIn in;		// these four should really be final
	public  AudioOut out;
	// fields for native code to use...
	private int lastError;
	private String lastErrorMsg;
	private int hwi,hwo;
	private boolean closed;

	/* Open a new audio device
	 */
	public Audio(int devID) throws AudioException {
		closed = false;
		try {
			synchronized (devices) {
				if ((devID < 0) || (devID >= numDev)) throw new AudioException("Illegal Device ID ("+devID+")");
				if (inUse(devID)) throw new AudioException("Device "+devID+" already in use");
				if (messageThread == null) {
					messageThread = new AudioMessageThread();
					synchronized (messageThread) {
						messageThread.start();
						try {
							messageThread.wait();
						} catch (InterruptedException e) {}
					}
				}
	//			System.out.println("Assigning device");
				devices[devID] = this;
				this.devID = devID;
				this.devName = deviceName[devID];
	//			System.out.println("Initializing");
				initN(messageThread.windowsID, debug);
				hwiDevices.put(new Integer(hwi), this);
				hwoDevices.put(new Integer(hwo), this);
	//			System.out.println("Finishing");
				checkError();
				in = new AudioIn(this);
				out = new AudioOut(this);
			}
		} catch (AudioException e) {
			try {
				close();
			} catch (Throwable e2) {
			}
			throw e;
		}
	}

	/* initialize this device (called by constructor) */
	private native void initN(int msgThreadID, boolean debug);
	
	/* check for errors on this device */
	protected void checkError() throws AudioException {
		if (lastError != 0)	{
			lastError = 0;
			throw new AudioException("Device "+devID+": ("+lastError+") "+lastErrorMsg);
		}
}

	/* close the device */
	public void close() throws AudioException {
		if (closed) return;
		synchronized (devices) {
			in.dev = out.dev = null;	// no one can use this device any more...
			hwiDevices.remove(new Integer(hwi));
			hwoDevices.remove(new Integer(hwo));
			closeN();
			devices[devID] = null;
			if (inUse() == 0) {
				messageThread.stop();
				messageThread = null;
			}
		}		
		closed = true;
		checkError();
	}
	/* native helper for closing device */
	private native void closeN();

	/* gc must close device properly */
	public void finalize() {
		try {
			close();
		} catch (AudioException e) {}
	}

}

